require 'uu/applog/application_log/application_log_record_attributes'
require 'uu/applog/application_log/severity'
require 'uu/os/uesuri'
require 'uu/os/uesuri_builder'
require 'time'
require 'log4r/MDC'
require 'log4r/outputter/outputter'
require 'thread'
require 'socket'

module UU
  module AppLog

    # Log4r outputter class for logger use with remote logging service application log by client api [UU::OS::AppLog::ApplicationLog].
    class Outputter < Log4r::Outputter

      # AppLog message queue
      # @private
      @@msg_queue ||= Queue.new

      # Asynchronous AppLog writer
      # @private
      @@worker ||= Thread.new(@@msg_queue) do |msg_queue|
        loop do
          if !msg_queue.empty?
            record = msg_queue.pop
            begin
              # Use auth token from thread where logging event was invoked
              # (do not call "login" to avoid remote token evaluation) 
              UU::OS::Security::Session.send(:put_authn_token, record[:token]) unless record[:token].nil?
              UU::AppLog::ApplicationLog.add_record(record[:applog_uri], record[:message], record[:outputter])
            rescue Exception => e
              Log4r::Logger.log_internal {"#{record[:outputter].class} #{record[:outputter].name}: Remote logging ends with exception: Exception message=#{e.message}; stacktrace=#{e.backtrace}"}
            ensure
              UU::OS::Security::Session.logout(true)
            end
          end
          # Prevent high CPU load on idling
          sleep(0.01)
        end
      end

      # UESURI of the specified application log.
      # @param [String, UU::OS::UESURI]
      # @return [UU::OS::UESURI]
      attr_accessor :log_uri

      # @private
      attr_accessor :uu_applog_auth_token

      # Tabbed line separator
      TABBED_LINE_SEPARATOR = "\n\t"
      # Separator of exception part in data field of [Log4r::LogEvent]. If we want store exception with stack trace to application log separated from logged message we have to use separator in data field because Log4r doesn't support in [Log4r::LogEvent] separated exception.
      # This separator is e.g. used by Log4rPlugin of [UU::OS::Logger].
      EXCEPTION_SEPARATOR = "#{TABBED_LINE_SEPARATOR}@@exception@@#{TABBED_LINE_SEPARATOR}"
      # key to MDC to get application log uri
      APPLOG_URI_KEY = 'uu-applogURI'
      # key to MDC to get application uri
      APPLICATION_URI_KEY = 'uu-appURI'
      # default application log code for generating default application log uri from application uri
      DEFAULT_APPLOG_CODE = 'APPLOG'

      # Creates new instance of Outputter.
      #
      # @param [String] _name the name of outputter
      # @param [Hash] hash map containing application log uri [String, UU::OS::UESURI] under the key +loguri+ or +log_uri+ or +logURI+
      def initialize(_name, hash={})
        super(_name, hash)
        uri = (hash[:loguri] or hash['loguri'] or hash[:log_uri] or hash[:logURI])
        if uri.nil? || uri.strip.empty?
          Log4r::Logger.log_internal {"#{self.class} #{self.name}: Application log uri is not defined in initialization. Application log URI can be defined or generated from application URI also dynamically by getting it from Log4r::MDC."}
        else
          @log_uri = UU::OS::UESURI.new(uri)
        end
      end

      # Method for outputter reconfiguration.
      # @private
      def reconfigure(options={})
        @log_uri = options[:appLogUri] if !options[:appLogUri].nil?
      end

      # This method handles all log events passed to a Outputter.
      # @param [Log4r::LogEvent] logevent wraps all data associated with a logging statement
      def canonical_log(logevent)
        require 'uu/applog/application_log' # the require must be called here to prevent cyclic dependency (which occurs if this require is at the top of the file)
        if loop_detected?
          Log4r::Logger.log_internal {"#{self.class} #{self.name}: Logging loop detected, because some component in the logging process called logger. Data being logged: logger name=#{logevent.fullname}, level=#{logevent.level}, message=#{logevent.data}, tracer=#{logevent.tracer}"}
          end_loop!
          return
        end
        begin_loop!
        begin
          applog_uri = log_uri
          if(applog_uri.nil?)
            Log4r::Logger.log_internal {"#{self.class} #{self.name}: Cannot log without UESURI of application log. First set log_uri parameter in log4r.xml or set on Log4r::MDC uu-appURI parameter or uu-applogURI parameter."}
            return
          end
          record = {}
          record[:applog_uri] = applog_uri
          record[:message] = create_message(logevent)
          record[:outputter] = self
          record[:token] = UU::OS::Security::Session.send(:get_authn_token)
          # Push message to queue (it will be written asynchronously).
          @@msg_queue.push(record)
        rescue Exception => e
          Log4r::Logger.log_internal {"#{self.class} #{self.name}: Remote logging ends with exception: Exception message=#{e.message}; stacktrace=#{e.backtrace}"}
        ensure
          end_loop!
        end
      end

      def log_uri
        return @log_uri unless @log_uri.nil?
        return get_mdc_log_uri
      end

      def log_uri=(val)
        if val
          @log_uri = UU::OS::UESURI.new(val)
        else
          @log_uri = nil
        end
      end

      private

      # Collect all logging data to DTO.
      # @return [UU::AppLog::ApplicationLog::ApplicationLogAddRecord] logevent
      def create_message(logevent)
        level = Log4r::LNAMES[logevent.level]
        message = UU::AppLog::ApplicationLog::ApplicationLogAddRecord.new
        #if message has exception and is separated by [EXCEPTION_SEPARATOR] as if created by UU::OS::Logger::Log4rPlugin.format_message
        #we can set message part and exception part. This have to be done because Log4r::LogEvent does not support exception.
        if logevent.data.include?(EXCEPTION_SEPARATOR)
          msgsplit = logevent.data.split EXCEPTION_SEPARATOR
          message.message = msgsplit[0]
          message.exception = msgsplit[1]
        else
          message.message = logevent.data
        end
        message.logger_code = logevent.fullname
        message.severity = from_log4r_level level
        message.creation_time = Time.now
        message.mdc = mdc_clone
        message.thread_id = Thread.current.to_s
        message.thread_name = Thread.current[:name]
        message.node_id = Socket.gethostname[0..255]
        message.node_ip = Socket.ip_address_list.map{|s| s.inspect}.join(',')[0..255]
        message
      end

      # Converts severity levels of log4r into application log severity levels.
      #
      # @param [String] level Severity level of log4r
      # @return [String] Severity level of application log
      def from_log4r_level(level)
        case level
        when Log4r::ERROR
          ApplicationLog::Severity::ERROR
        when Log4r::WARN
          ApplicationLog::Severity::WARNING
        when Log4r::INFO
          ApplicationLog::Severity::INFO
        when Log4r::DEBUG
          ApplicationLog::Severity::DEBUG
        when 'ERROR'
          ApplicationLog::Severity::ERROR
        when 'WARN'
          ApplicationLog::Severity::WARNING
        when 'INFO'
          ApplicationLog::Severity::INFO
        when 'DEBUG'
          ApplicationLog::Severity::DEBUG
        else
          nil
        end
      end

      # Sets a mark to prevent logging loop.
      def begin_loop!
        Log4r::MDC.put(:uu_applog_outputter_loop_detection, 'looped')
      end

      # Clears a mark previously set to prevent logging loop.
      def end_loop!
        Log4r::MDC.put(:uu_applog_outputter_loop_detection, nil)
      end

      # Checks if a logging loop has been detected.
      def loop_detected?
        if (Log4r::MDC.get(:uu_applog_outputter_loop_detection) == nil)
          return false
        end
        if (Log4r::MDC.get(:uu_applog_outputter_loop_detection).empty?)
          return false
        end
        return true
      end

      # Clone mdc for sending to application log and delete control parameters
      def mdc_clone
        result = Log4r::MDC.get_context
        result.delete :uu_applog_outputter_loop_detection
        result.delete APPLOG_URI_KEY
        result.empty? ? nil : result
      end

      # find application log uri on MDC or create default log uri from application uri stored on mdc
      def get_mdc_log_uri
        applog_uri = Log4r::MDC.get(APPLOG_URI_KEY)
        ex1 = nil
        begin
          if (!applog_uri.nil?) && (!applog_uri.strip.empty?)
            return UU::OS::UESURI.new(applog_uri)
          else
            ex1 = 'Missing application log URI stored on MDC.'
          end
        rescue Exception => e
          ex1 = e.message
        end
        app_uri = Log4r::MDC.get(APPLICATION_URI_KEY)
        ex2 = nil
        begin
          if (!app_uri.nil?) && (!app_uri.strip.empty?)
            uri = UU::OS::UESURI.new app_uri
            # we need code of application otherwise cannot continue
            application_code = uri.artifact_code
            if !application_code.nil? && !application_code.empty?
              #create default application log code
              applog_code = "#{application_code}/#{DEFAULT_APPLOG_CODE}"
              uri_builder = UU::OS::UESURIBuilder.new
              uri_builder.set_territory_code uri.territory_code
              uri_builder.set_territory_id uri.territory_id
              uri_builder.set_artifact_code applog_code
              return uri_builder.to_uesuri
            else
              ex2 = "Missing application code in application URI stored on MDC. URI = #{app_uri}"
            end
          else
            ex2 = 'Missing application URI stored on MDC.'
          end
        rescue Exception => e
          ex2 = e.message
        ensure
          Log4r::Logger.log_internal {"#{self.class} #{self.name}: Trying get URI of application log or application URI to generate default application log URI from Log4r::MDC by key=uu-applogURI or by key=uu-appURI failed. Application log uri parse exception message=#{ex1} or application uri parse exception message=#{ex2}."} unless ex1.nil? || ex2.nil?
          if !ex1.nil? && !ex2.nil?
            return nil
          end
        end
      end
    end
  end
end
